eureqa_colors = {
    "WHITE": "#fff",
    "GREEN": "#72bb3a",
    "ORANGE": "#e77219",
    "YELLOW": "#FCD051",
    "RED": "#c91f3e",
    "GOLD": "#edb619",
    "PRIMARY_BLUE": "#00a4c9",
    "SECONDARY_BLUE": "#6296C3",
    "LINK_BLUE": "#006699",
    "LIGHT_GREY": "#ddd",
    "MEDIUM_GREY": "#aaa",
    "DARK_GREY": "#5f605f",
    "ICON_GREY": "#5f605f",
    "DEFAULT_FONT": "#333333",
    "LIGHT_FONT": "#aaa",
    "BACKGROUND_GREY": "#f4f4f4",
    "PRIMARY_BUTTON": "#e77219",
    "SECONDARY_BUTTON": "#777",
    "BOX_SHADOW": "0px 2px 3px 0px #ddd",
    "BORDER_LIGHT": "1px solid #ddd",
    "BORDER_MEDIUM": "1px solid #aaa",
    "BORDER_DOTTED": "1px dotted #aaa",
    "BORDER_HIGHLIGHT": "1px solid #00a4c9",
    "TABLE_ROW_SELECTED": "1px double #00a4c9"
}


# Copyright (c) 2016, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import uuid
from eureqa.utils.jsonrest import _JsonREST

# Python Metaclass that automatically tracks all known types of Components
# If you're in this file, you probably want to skip ahead to 'class Component()' below.
# The following code has a very specific use case.
class _ComponentRegistrar(type):
    """ For internal use only. """

    # For those not familiar with Python metaclasses and __new__:
    #
    # __init__() runs when an *instance* is created.
    # __new__() runs when a *class* is created.
    #
    # In the common case (unless you're creating classes on the fly), a class
    # is "constructed" in this sense when Python first opens and parses the
    # file containing this class.
    #
    # The "Component" class represents a common base class for a
    # collection of serializable objects.
    # The serialized representation of the class contains a 'type' string
    # that indicates what type of thing has been serialized; therefore,
    # which Python class should be used to represent it.
    # In order to de-serialize such objects, 'Component' must therefore
    # know about all of its subclasses that represent specific
    # types of objects, as well as which class represents which 'type'
    # string.
    # Rather than write and maintain a list, this code makes the Python
    # interpreter generate the list for us.

    def __new__(type_, name, bases, dict_):
        cls = type.__new__(type_, name, bases, dict_)
        assert hasattr(cls, "_component_type_str"), \
            "New Analysis Component classes must define a '_component_type' field"
        cls._register(cls._component_type_str)
        return cls


class _Component(_JsonREST):
    """
    Component base class.
    Implements common logic shared by all components.
    Do not construct this class directly; instead,
    construct the specific component that you want.

    Notes for component implementors:  (If your new class's sub-Components aren't showing up properly, read on!)

    The backend first sees a Component when the first of the following happens:
      * The Component is constructed with an Analysis passed into the constructor as an argument
      * The Component is added to an Item that is already attached to an Analysis
      * The Component is added to an Item that is not already attached to an Analysis, and then the Item is attached to an Analysis
      * The Component is added to a Table and that Table's data is uploaded
    If more than one of these things happens in the lifecycle of a Component, or if one happens repeatedly,
    only the very first occurrence counts.
    Call this first event "Association".

    Before Association, each individual Component subclass is responsible for tracking their sub-Components (if any).
    At Association, "Component._associate()" is called; each Component that holds sub-Components must override this
    method to also call "_associate()" on all of their sub-Components.
    After Association (and *NOT* before it), each Component is responsible for immediately calling "_associate()"
    on any sub-Component that is added to it.
    """

    __metaclass__ = _ComponentRegistrar

    def clone(self):
        # Make a copy of ourselves
        body = self._to_json()

        # Disassociate the copy from the current Analysis
        if "analysis_id" in body:
            del body["analysis_id"]
        if "component_id" in body:
            del body["component_id"]

        # Instantiate a new Item with this state
        return _Component._construct_from_json(body)

    def delete(self):
        """ Remove this Component from the server. """
        # Subclasses may override this to do more elaborate clean-up.
        self._delete()

    def __init__(self, _analysis=None, _component_id=None, _component_type=None):
        """ For internal use only. """
        if _component_id is not None:
            self._component_id = _component_id

        if _component_type is not None:
            self._component_type = _component_type
        else:
            self._component_type = self._component_type_str

        if _analysis:
            self._associate(_analysis)

    _component_type_str = "GENERIC_COMPONENT"
    _registered_types = {}

    @classmethod
    def _register(cls, component_type):
        cls._registered_types[component_type] = cls

    def _populate_eureqa(self):
        if not hasattr(self, "_analysis"):
            # Can't populate Eureqa yet
            return

        self._eureqa = self._analysis._eureqa


    # Override "_construct_from_json()" to use the appropriate subclass if possible
    @classmethod
    def _construct_from_json(cls, body, *args, **kwargs):

        # If we're given an analysis, pass it in after
        # we're done populating the object so that we don't make
        # extraneous HTTP requests populating partial object state.
        analysis = None
        if "_analysis" in kwargs:
            analysis = kwargs["_analysis"]
            del kwargs["_analysis"]

        ret = cls._registered_types.get(body.get("component_type"), _Component)(*args, **kwargs)
        ret._from_json(body)

        if analysis is not None:
            ret._associate(analysis)

        ret._populate_eureqa()
        return ret

    # Override "_get_all_from_self()" to associate all newly-constructed components
    # with our Analysis (if any)
    def _get_all_from_self(self):
        """
        Fetch a list of all objects in the same directory as this object.
        :return: list() of objects of the current type
        """
        comps = super(_Component, self)._get_all_from_self()

        if hasattr(self, "_analysis"):
            for comp in comps:
                comp._associate(self._analysis)

        return comps

    def _object_endpoint(self):
        return '/analysis/%s/components/%s' % (self._analysis._id, self._component_id)

    def _directory_endpoint(self):
        return '/analysis/%s/components' % (self._analysis._id)

    def _fields(self):
        return [ "analysis_id", "component_id", "component_type" ]

    def _to_json(self):
        resp = super(_Component, self)._to_json()

        if hasattr(self, "_analysis"):
            resp["analysis_id"] = self._analysis._id

        return resp

    def _from_json(self, body):
        if hasattr(self, "_analysis") and body.get("analysis_id"):
            assert body.get("analysis_id") == self._analysis._id, \
                "_from_json() can't de-serialize an Item that belongs to a different Analysis"

        super(_Component, self)._from_json(body)

    def _associate(self, analysis):
        self._analysis = analysis
        self._analysis_id = analysis._id
        _JsonREST.__init__(self, analysis._eureqa)
        self._populate_eureqa()
        if not hasattr(self, "_component_id"):
            # Get an ID and whatnot only if we don't already have one.
            self._create()

    def _associate_with_table(self, analysis):
        # The base class implementation of this method only generates UUID but
        # does not add them to the actual analysis. Because most components
        # are directly inlined into the table and some involve extra actions (e.g. uploading a file)
        # It is up to a particular component to override this method and provide more
        # specific implementation.
        if not hasattr(self, "_component_id"):
            self._component_id = str(uuid.uuid4())
        return [self._to_json()]

class _TwoVariablePlotComponent(_Component):
    _component_type_str = "TwoVariablePlot helper mixin (not a complete type)"


    def __init__(self, _analysis=None, _component_id=None, _component_type=None):

        super(_TwoVariablePlotComponent, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

        default_x_label = getattr(self, '_x_var', 'the x var')
        default_y_label = getattr(self, '_y_var', 'the y var')

        DEFAULTS = {'BOX_PLOT':             {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                             'default_format': { 'x': 'g',   'y': '.2s' }},

                    'DOUBLE_HISTOGRAM_PLOT':{'default_label' : { 'x': [default_x_label, default_y_label], 'y': 'number of values' },
                                             'default_format': { 'x': 'g', 'y': '.3s' }},

                    'SCATTER_PLOT':         {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                             'default_format': { 'x': '.3s', 'y': '.3s' }},

                    'BINNED_MEAN_PLOT':     {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                             'default_format': { 'x': '.3s', 'y': '.2f' }}}



        default_values = DEFAULTS.get(self._component_type,
                                      {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                       'default_format': { 'x': '.3s', 'y': '.3s' }})

        if not hasattr(self,'_axisLabels'):
            self._axisLabels = default_values['default_label']
        if not hasattr(self,'_labelFormat'):
            self._labelFormat = default_values['default_format']


    class XYMap(object):
        """ A named tuple with keys 'x' and 'y'

        :param TwoVariablePlotCard tvp: The TwoVariablePlotCard object.
        :param dict dic: X and Y parameters as dictionary.
        """

        def __init__(self, parent_component, body):
            self._parent_component = parent_component
            self._body = body

        @property
        def x(self):
            """ X value """
            return self._body['x']

        @x.setter
        def x(self, val):
            self._body['x'] = val
            self._update()

        @property
        def y(self):
            """ Y value """
            return self._body['y']

        @y.setter
        def y(self, val):
            self._body['y'] = val
            self._update()

        def _update(self):
            self._parent_component._update()



# Copyright (c) 2016, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime

import json

class ProgressUpdate:
    """Represents an analysis template execution progress update on the server.
 
    :param dict body: Class metadata as dictionary

    :var str message: The the message about the execution's status.
    :var datetime time_stamp: The time the progress update was created.
    """

    def __init__(self, body):
        """For internal use only"""

        self.message = body['message']
        self._time_stamp = body['time_stamp']
        self._body = body

    @property
    def time_stamp(self):
        return datetime.strptime(self._time_stamp, '%Y-%b-%d %H:%M:%S.%f')

    def _to_json(self):
        body = {
            'message': self.message,
            'time_stamp': self.time_stamp.strftime('%Y-%b-%d %H:%M:%S.%f')
        }
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)



# Copyright (c) 2016, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from error_metric import ErrorMetrics

import json

class Solution:
    """Represents one of the solutions found by the server for a particular search. Can be obtained from
    one of the methods in :py:class:`~eureqa.search.Search`. Some solutions may be available even if the
    search has not finished running.

    :var str `~eureqa.solution.Solution.target`: The target variable.
    :var str `~eureqa.solution.Solution.model`: The model expression.
    :var int `~eureqa.solution.Solution.complexity`: The model complexity based on the complexity weights.
    :var bool `~eureqa.solution.Solution.is_best`:
        An indicator whether the solution is considered to be the best based on its complexity and precision.
        The system can pick only one solution as best.
    :var bool `~eureqa.solution.Solution.is_most_accurate`:
        An indicator whether the solution is the most accurate for the optimized error metric.
        The system can pick only one solution as most accurate.
    :var str `~eureqa.solution.Solution.optimized_error_metric`:
        The error metric for which the solution was optimized.
        It is one of the error metrics from :mod:`~eureqa.error_metric`.
    :var float `~eureqa.solution.Solution.optimized_error_metric_value`: The value of the optimized error metric.
    :var ~eureqa.search.Search `~eureqa.solution.Solution.search`: The search object to which this solution belongs.
    """

    def __init__(self, body, eureqa, search=None):
        """For internal use only.
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._id = body['solution_id']
        self._search_id = body['search_id']
        self._datasource_id = body['datasource_id']
        self.target = body['target_variable']
        self.model = body['model_string']
        self.complexity = body['model_complexity']
        self.is_best = body['solution_type']['best']
        self.is_most_accurate = body['solution_type']['most_accurate']
        self._error_metrics = {x['metric_name']: x['metric_value'] for x in body['error_metric_values']}
        self.optimized_error_metric = body['model_optimized_error_metric']
        self.optimized_error_metric_value = body['model_optimized_error_metric_value']
        self._model_fitness = body['model_fitness']
        self._model_aic_parameters = body.get('model_aic_parameters')
        self.terms = [t['term_string'] for t in body['terms']]
        self._body = body
        self._eureqa = eureqa
        self._search_object = search

    def _to_json(self):
        body = {
            'target': self.target,
            'model': self.model,
            'complexity': self.complexity,
            'is_best': self.is_best,
            'is_most_accurate': self.is_most_accurate,
            'optimized_error_metric': self.optimized_error_metric,
            'optimized_error_metric_value': self.optimized_error_metric_value,
            'terms': self.terms
        }
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    def get_error_metric_value(self, name):
        """
        Returns the value for the specified error metric.

        :param str name: One of the error metrics from :mod:`~eureqa.error_metric`.
        :rtype: float
        """

        return self._error_metrics[name]

    def get_all_series_error_metrics(self, data_split='all'):
        """
        Returns the metric values for each series in the datasource.

        :param str data_split: For internal use only.

        :return: list of :class:`~eureqa.error_metric.ErrorMetrics` objects.
        :rtype: list
        """

        result = self.search._eureqa._session.execute('/fxp/datasources/%s/searches/%s/solutions/%s/series_metrics?data_type=%s'% (self.search._data_source_id, self.search._id, self._id, data_split), 'GET')
        metrics_list = []
        for x in result:
            metrics = ErrorMetrics()
            metrics._from_json(x)
            metrics_list.append(metrics)
        return metrics_list

    def get_single_series_error_metrics(self, series_index, data_split='all'):
        """
        Returns the metric values for the specified series in the datasource.

        :param int series_index: The series to compute the metrics on
        :param str data_split: For internal use only.
        :rtype: ErrorMetrics
        """

        result = self.search._eureqa._session.execute('/fxp/datasources/%s/searches/%s/solutions/%s/series_metrics/%s?data_type=%s'% (self.search._data_source_id, self.search._id, self._id, series_index, data_split), 'GET')
        metrics = ErrorMetrics()
        metrics._from_json(result)
        return metrics

    @property
    def search(self):
        if not self._search_object:
            self._search_object = self._eureqa._get_search_by_search_id(self._datasource_id, self._search_id)
        return self._search_object



# Copyright (c) 2016, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class ObjectStore:
    """For internal use only. Used to get/put/delete objects from the object store.
    
    :param Eureqa eureqa: a eureqa connection
    """

    def __init__(self, eureqa):
        self._session = eureqa._session

    _uri_root = '/objstore'

    @classmethod
    def get_uri_parts(cls, uri):
        """Given a uri of the form /this/is/a/uri, return a list of the uri components.

        :param obj cls: Classmethod object
        :param str uri: The URI to process.
        :return: a list of the uri components
        :rtype: list
        """

        return uri.split('/')

    @classmethod
    def construct_uri(cls, bucket, key=None):
        """Given a bucket and optional key, construct the objectstore uri
        associated with those values.

        :param obj cls: Classmethod object
        :param str bucket: Objectstore logical container
        :param str key: Objectstore key
        :return: the objectstore uri associated with the specified bucket/key
        :rtype: str
        """

        parts = [cls._uri_root, bucket]
        if key is not None: parts.append(key)
        return '/'.join(parts)

    @classmethod
    def construct_uri_full(cls, organization, bucket, key=None, download=False):
        """Given an organization, bucket and optional key, construct the absolute uri
        associated with those values, starting with '/api'.
        This uri will be accessible from a Web browser.  It may not work from Python.

        :param obj cls: Classmethod object
        :param str organization: The logical organization containing the data.
        :param str bucket: Objectstore logical container
        :param str key: Objectstore key
        :param bool download: Whether to enable download from this URI
        :return: the full objectstore uri associated with the specified bucket/key
        :rtype: str
        """

        uri = cls.construct_uri(bucket, key)
        uri_full = '/api/' + organization + uri
        if download: uri_full += '?download=true'
        return uri_full

    @classmethod
    def construct_python_uri_full(cls, organization, bucket, key=None, download=False):
        """Given an organization, bucket and optional key, construct the absolute uri
        associated with those values, starting with '/api/v2'.
        This uri will be accessible from a python.  It may not work from a Web browser.

        :param obj cls: Classmethod object
        :param str organization: The logical organization containing the data.
        :param str bucket: Objectstore logical container
        :param str key: Objectstore key
        :param bool download: Whether to enable download from this URI
        :return: the full objectstore uri associated with the specified bucket/key
        :rtype: str
        """
        uri = cls.construct_uri(bucket, key)
        uri_full = '/api/v2/' + organization + uri
        if download: uri_full += '?download=true'
        return uri_full

    def get(self, bucket, key=None, filepath=None):
        """Get an object from a bucket/key in the objectstore.

        If no key is specified, the response will be a list of keys in the
        specified bucket

        If filename is specified, result is also saved to file.

        :param str bucket: Objectstore logical container
        :param str key: Objectstore key
        :param str filepath: The (optional) key for this data URI 
        :return: the contents of the bucket/key
        :rtype: str
        """

        uri = self.construct_uri(bucket, key)

        response = self._session.execute(uri, 'GET', raw_returnfile=filepath)
        return response

    def put(self, bucket, key=None, raw_data=None, filepath=None):
        """Put contents into a bucket/key.

        If a key is specified, we'll overwrite the stored value for the
        bucket/key. Otherwise we'll create a new key in the bucket.

        :param str bucket: Objectstore logical container
        :param str key: Objectstore key
        :param bool raw_data: The raw data if provided
        :param str filepath: The filepath if desired
        :return: the response from the objectstore
        :rtype: str
        """

        uri = self.construct_uri(bucket, key)
        files = {'file':open(filepath, 'rb')} if filepath else None
        response = self._session.execute(uri, 'POST', raw_data=raw_data, files=files)
        if files: files['file'].close()
        return response

    def delete(self, bucket, key=None):
        """Delete contents of a bucket or key.

        If no key is specified, we'll delete the entire bucket.

        :param str bucket: Objectstore logical container
        :param str key: Objectstore key
        """

        uri = self.construct_uri(bucket, key)
        response = self._session.execute(uri, 'DELETE')
        return response




# Copyright (c) 2016, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from search import Search
from variable_details import VariableDetails
from utils import *
from utils.jsonrest import _JsonREST

from session import Http404Exception

import base64
import json
import urllib

import warnings

class DataSource(_JsonREST):
    """Acts as an interface to a data source on the server.

    DataSources can be created by calling :py:meth:`~eureqa.eureqa.Eureqa.create_data_source`
    or an existing one can be retreived with :py:meth:`~eureqa.eureqa.Eureqa.get_data_source`

    :var str `~eureqa.data_source.DataSource.name`: The data source name.
    :var int `~eureqa.data_source.DataSource.number_columns`: The number of columns (variables) in the data source.
    :var int `~eureqa.data_source.DataSource.number_rows`: The number of rows in the data source.
    :var int `~eureqa.data_source.DataSource.number_series`: The number of series (chunks of rows) in the data source.
    """

    def __init__(self, eureqa, datasource_name=None, hidden=False, datasource_id=None):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        super(DataSource, self).__init__(eureqa)

        if datasource_id is not None:
            self._datasource_id = datasource_id
        if datasource_name is not None:
            self._datasource_name = datasource_name
        if hidden is not None:
            self._hidden = hidden
        self._column_count = 0
        self._row_count = 0
        self._series_count = 0
        self._series_sizes = []
        self._file_name = ''
        self._file_size = 0
        self._file_uploaded_user = ''
        self._file_uploaded_date = ''
        self._file_objstore_uri = ''

    def _directory_endpoint(self):
        return "/fxp/datasources"
    def _object_endpoint(self):
        return "/fxp/datasources/%s" % (self._datasource_id)
    def _fields(self):
        return ["datasource_id", "datasource_name", "row_count", "column_count", "series_count", "series_sizes", "file_name", "file_size", "file_uploaded_user", "file_uploaded_date", "file_objstore_uri", "hidden"]

    @property
    def name(self):
        return self._datasource_name
    @name.setter
    def name(self, val):
        self._datasource_name = val
        self._update()

    @property
    def number_columns(self):
        return self._column_count

    @property
    def number_rows(self):
        return self._row_count

    @property
    def number_series(self):
        return self._series_count

    @property
    def _data_source_id(self):
        return self._datasource_id

    @property
    def _data_file_name(self):
        return self._file_name

    @property
    def _data_file_size(self):
        return self._file_size

    @property
    def _data_file_uploaded_user(self):
        return self._file_uploaded_user

    @property
    def _data_file_uploaded_date(self):
        return self._file_uploaded_date

    @property
    def _data_file_objstore_uri(self):
        return self._file_objstore_uri

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    def delete(self):
        """Deletes the data source from the server.

        :raise Exception: If the data source is already deleted.
        """
        self._delete()

    def _associate_data_file(self, file_or_path):
        assert hasattr(self, '_datasource_id'), "Error: Cannot associate data to DataSource with no datasource_id. Make sure the DataSource has been posted to the server."
        if isinstance(file_or_path, basestring):
            f = open(file_or_path, 'rb')
            file_path = file_or_path
        else:
            f = file_or_path
            file_path = getattr(file_or_path, 'name', str(file_or_path))

        try:
            self._eureqa._session.execute('/fxp/datasources/%s/data' % self._datasource_id, 'POST', files={'file': f})
        finally:
            if isinstance(file_or_path, basestring):
                f.close()
        self._get_self()

    def download_data_file(self, file_path):
        """Downloads the originally uploaded data file from the server.

        :param str file_path: the filepath at which to save the data

        """
        if not self._data_file_objstore_uri: raise Exception("Datasource %s is too old. The original data can only be retrieved from datasources made with newer versions of Eureqa" % (self.name))
        result = self._eureqa._session.execute(self._data_file_objstore_uri, 'GET', raw_returnfile=file_path)

    def get_variables(self):
        """Retrieves from the server a list of variables in a data set.

        :return:
            A list of the same variables as visible in Eureqa UI.
            Including all derived variables.
        :rtype: list of str
        """
        endpoint = '/fxp/datasources/%s/variables?sort=[{"key":"index"}]' % self._datasource_id
        self._eureqa._session.report_progress('Getting variable details for datasource: \'%s\'.' % self.name)
        body = self._eureqa._session.execute(endpoint, 'GET')
        return [x['variable_name'] for x in body]

    def get_searches(self):
        """Retrieves from the server a list of searches associated with the data source.

        :return: The list of all searches associated with the data source.
        :rtype: list of :class:`~eureqa.search.Search`
        """

        endpoint = '/fxp/datasources/%s/searches' % self._datasource_id
        self._eureqa._session.report_progress('Getting searches for datasource: \'%s\'.' % self.name)
        body = self._eureqa._session.execute(endpoint, 'GET')
        return [Search(x, self._eureqa, self) for x in body]

    def create_search(self, search_settings, _hidden = False):
        """Creates a new search with settings from a :any:`SearchSettings` object.

        :param SearchSettings search_settings: the settings for creating a new search.
        :return: A :class:`~eureqa.search.Search` object which represents a newly create search on the server.
        :rtype: ~eureqa.search.Search
        """

        endpoint = "/fxp/datasources/%s/searches" % self._datasource_id
        body = search_settings._to_json()
        body['hidden'] = _hidden
        self._eureqa._session.report_progress('Creating search for datasource: \'%s\'.' % self.name)
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        search_id = result['search_id']
        return self._eureqa._get_search_by_search_id(self._datasource_id, search_id)

    def evaluate_expression(self, expressions, _data_split='all'):
        warnings.warn("This function has been deprecated.  Please use `Eureqa.evaluate_expression()` instead.", DeprecationWarning)
        return self._eureqa.evaluate_expression(self, expressions, _data_split=_data_split)

    def _compute_series_indicator(self):
        """For internal use only.

        Compute an indicator column which indicates the series which each row belongs to, as an integer from 0 to (num_series-1).

        This code was written with the assumption that it would be applied to the entire dataset, as opposed to the training or validation split.
        """
        indicator = []
        for i,series_size in enumerate(self._series_sizes):
            indicator += [i]*series_size
        return indicator

    def create_variable(self, expression, variable_name):
        """Adds a new variable to the data_source with values from evaluating the given expression.

        :param str expression: the expression to evaluate to fill in the values
        :param str variable_name: what to name the new variable
        """
        endpoint = '/fxp/datasources/%s/variables' % self._datasource_id
        body = {'datasource_id': self._datasource_id,
                'expression': expression,
                'variable_name': variable_name}
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        self.__dict__ = self._eureqa.get_data_source_by_id(self._datasource_id).__dict__

    def get_variable_details(self, variable_name):
        """Retrieves the details for the requested variable from the data_source.

        :param str variable_name: the name of the variable to get the details for
        :return:
            The object representing the variable details
        :rtype: VariableDetails
        """
        endpoint = '/fxp/datasources/%s/variables/%s' % (self._datasource_id, urllib.quote_plus(base64.b64encode('%s-_-%s' % (self._datasource_id, variable_name))))
        self._eureqa._session.report_progress('Getting variable details for datasource: \'%s\'.' % self.name)
        body = self._eureqa._session.execute(endpoint, 'GET')
        return VariableDetails(body, self)

    def get_variable(self, variable_name):
        warnings.warn("'get_variable()' function deprecated; please call as 'get_variable_details()' instead", DeprecationWarning)
        return self.get_variable_details(variable_name)



# Copyright (c) 2016, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
import os
from objectstore import ObjectStore

class Image:
    """Represents an image which has been uploaded to the server

    :param Eureqa eureqa: A eureqa connection.
    :param str image_filename: The name of the image file.
    :param str image_key: The identifying key of the image file in the objectstore image_uploads bucket.
    """

    def __init__(self, eureqa, image_filename, image_key):
        """Represents an image stored on the server.
        """

        self._eureqa = eureqa
        self._filename = image_filename
        self._key = image_key

    """For internal use only. ObjectStore bucket in which images are stored."""
    _image_bucket = 'image_uploads'

    @property
    def image_filename(self): return self._filename

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
               self._filename == other._filename and \
               self._key == other._key

    def __ne__(self, other): return not self.__eq__(other)

    def __str__(self):
        """Get the markup text used to embed this image in an analysis `~TextCard`.

        :return: markup text
        :rtype: str
        """

        # get fully qualified uri of the form '/api/v2/:org/objstore/image_uploads/:image_key'
        image_text = '![%s](%s)' % (self.image_filename, self._web_url)
        return image_text

    @property
    def _python_url(self):
        # get fully qualified uri of the form '/api/v2/:org/objstore/image_uploads/:image_key'
        return ObjectStore.construct_python_uri_full(self._eureqa._session.organization, self._image_bucket, self._key)

    @property
    def _web_url(self):
        # get fully qualified uri of the form '/api/:org/objstore/image_uploads/:image_key'
        return ObjectStore.construct_uri_full(self._eureqa._session.organization, self._image_bucket, self._key)

    @classmethod
    def upload_from_file(cls, eureqa, image_path):
        """Uploads an image from file.

        :param str cls: Classpath object. 
        :param Eureqa eureqa: A eureqa connection.
        :param str image_path: The path to an image file.
        :return: An Image object representing the newly uploaded image.
        :rtype: :class:`eureqa.utils.Image`
        """

        objectstore = ObjectStore(eureqa)

        filename = os.path.basename(image_path)
        response = objectstore.put(cls._image_bucket, filepath=image_path)
        key = response['key']
        return Image(eureqa, filename, key)

    def _download(self, image_path):
        """Download the image file to disk.

        Private for now to keep the public api simple

        :param str image_path: where to save the image file. Should be full filepath, not directory.
        """

        objectstore = ObjectStore(self._eureqa)
        response = objectstore.get(self._image_bucket, self._key, filepath=image_path)

    def delete(self):
        """Removes the image from the server. Caller is expected to throw away this object afterwards.
        """

        objectstore = ObjectStore(self._eureqa)
        response = objectstore.delete(self._image_bucket, self._key)

    @classmethod
    def _get_images_from_text(cls, eureqa, text):
        """Given some text, get a list of `~Image` objects representing all the images embedded in the text.

        Private for now to keep the public api simple

        :param str text: markup text which may contain references to images.
        :return: a list of image objects in the order in which they appear.
        :rtype: list of :class:`~eureqa.utils.Image`
        """

        # extract the file names and uris of any embedded images in the card text.
        # Image embeds are of the form ![Image Filename.xxx](/api/v2/:org/objstore/image_uploads/:image_key).
        # So this regex has two groups. First group captures the filename, and second captures the image
        # uri
        regex_result = re.findall('\!\[([^\[^\]]*)\]\(/api([^\(^\)]*)\)', text)
        # since regex has two groups, result is of the form [('filename1','uri1'),('filename2','uri2'),...]
        # we also extract the image keys from the image uris by getting the last part of the uri
        images = [Image(eureqa, filename, ObjectStore.get_uri_parts(uri)[-1]) for filename, uri in regex_result]
        return images



from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST


class Layout(_Component):
    """Generic layout within an Analysis

    A generic grid layout composed of rows and columns sized by
    fractions (`1`, `1/2`, `1/4`, etc). Layouts can be nested within
    other layouts to create any type of grid needed.

    For example::

       layout = Layout()
       layout = Layout(borders=True)                                       # create a layout where borders are displayed
       layout.add_component(HtmlBlock(html="<h1>Hybrid Performance</h1>")) # add an item that is an entire row
       layout.add_row()                                                    # start a new row. newly added components will be placed into the new row.
       layout.add_component(MagnitudeBar(value=-0.22), "1/3")              # add a magnutide bar on the left 1/3 of the screen
       layout.add_component(MagnitudeBar(value=-0.22), "1/3", True)        # add a magnutide bar with dark background enabled
       layout.add_component(HtmlBlock(html="Variables...."), "1/2")        # add a text panel in the next 1/2 of the screen
       analysis.create_card(layout)                                        # Add this Layout to an Analysis

    :param list rows: (optional) List of :class:`~eureqa.analysis.components.Layout.Row` objects to add to this class initially

    """

    class _GridItem(object):
        """ Internal implementation class """
        def __init__(self, component=None, grid_unit="1", dark_background=False, analysis=None):
            self.grid_unit = grid_unit
            self.dark_background = dark_background
            self.component = component
            if component is not None and hasattr(component, "_component_id"):
                self._component_id = component._component_id
            self._analysis = analysis

        def _to_json(self):
            return {
                "grid_unit": self.grid_unit,
                "dark_background": self.dark_background,
                "component_id": self.component._component_id
            }

        def _from_json(self, body):
            if "grid_unit" in body:
                self.grid_unit = body["grid_unit"]
            if "dark_background" in body:
                self.dark_background = body["dark_background"]
            if "component_id" in body:
                self._component_id = body["component_id"]
                if self._analysis is not None:
                    self.component = _Component(_analysis=self._analysis, _component_id=body["component_id"])._get_self()

        def _associate(self, analysis):
            if self.component is None:
                self.component = _Component(_analysis=analysis, _component_id=getattr(self, "_component_id", None))._get_self()
            self.component._associate(analysis)

        def _associate_with_table(self, analysis):
            if self.component is None:
                self.component = _Component(_analysis=analysis, _component_id=getattr(self, "_component_id", None))._get_self()
            return self.component._associate_with_table(analysis)

    class _Row(object):
        """ Internal implementation class """
        def __init__(self, analysis=None, grid_items=None, ):
            self.grid_items = grid_items if grid_items else []
            self._analysis = analysis

        def create_card(self, component, grid_unit="1", dark_background=False):
            self.grid_items.append(Layout._GridItem(component, grid_unit, dark_background, self._analysis))
            return self.grid_items[-1]

        def _to_json(self):
            return [item._to_json() for item in self.grid_items]

        def _from_json(self, val):
            self.grid_items = []
            for item_json in val:
                item = Layout._GridItem(analysis=self._analysis)
                item._from_json(item_json)
                self.grid_items.append(item)

        def _associate(self, analysis):
            for item in self.grid_items:
                item._associate(analysis)

        def _associate_with_table(self, analysis):
            return [item._associate_with_table(analysis) for item in self.grid_items]

    _component_type_str = 'GENERIC_LAYOUT'

    def __init__(self, rows=None, borders=False, _analysis=None, _component_id=None, _component_type=None):
        if rows is not None:
            self.rows = rows
        else:
            self.rows = []
        self._borders = borders

        super(Layout, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def _current_row(self):
        # Break with convention:
        # 'self.rows' is not a property that points to 'self._rows
        if len(self.rows) == 0:
            self.rows = [ Layout._Row(analysis=getattr(self, "_analysis", None)) ]

        return self.rows[-1]

    @property
    def borders(self):
        return self._borders

    def add_row(self):
        self.rows.append(Layout._Row(analysis=getattr(self, "_analysis", None)))
        return self.rows[-1]

    def add_component(self, component, grid_unit="1", dark_background=False):
        return self._current_row.create_card(component, grid_unit, dark_background)

    def delete(self):
        for row in self.rows:
            for item in row.grid_items:
                item.component.delete()
        super(Layout, self).delete()

    def _to_json(self):
        ret = super(Layout, self)._to_json()
        ret["rows"] = [row._to_json() for row in self.rows]
        return ret

    def _from_json(self, body):
        super(Layout, self)._from_json(body)
        self.rows = []
        for row_json in body.get("rows", []):
            row = Layout._Row(analysis=getattr(self, "_analysis", None))
            row._from_json(row_json)
            self.rows.append(row)

    def _associate(self, analysis):
        super(Layout, self)._associate(analysis)
        for row in self.rows:
            row._associate(analysis)

    def _fields(self):
        # 'rows' is handled in _to_json
        return super(Layout, self)._fields() + [ "borders" ]

    def _associate(self, analysis):
        for row in self.rows:
            row._associate(analysis)
        super(Layout, self)._associate(analysis)

    def _associate_with_table(self, analysis):
        components_json = []
        for row in self.rows:
            components_json += row._associate_with_table(analysis)
        return super(Layout, self)._associate_with_table(analysis) + components_json
